/******************************************************************************/
/* SlunkCrypt, by LoRd_MuldeR <MuldeR2@GMX.de>                                */
/* This work has been released under the CC0 1.0 Universal license!           */
/******************************************************************************/

#if defined(_MSC_VER) && !defined(_DLL)
#define PTW32_STATIC_LIB 1
#endif

/* Internal */
#include "slunkcrypt.h"
#include "thread.h"

/* CRT */
#include <stdlib.h>

/* PThread */
#include <pthread.h>
#ifdef __unix__
#include <sys/sysinfo.h>
#endif

/* States */
#define THRD_STATE_IDLE 0
#define THRD_STATE_WORK 1
#define THRD_STATE_EXIT 2

// ==========================================================================
// Data types
// ==========================================================================

typedef struct
{
	thrdpl_worker_t worker;
	void *context;
	size_t length;
	const uint8_t *input;
	uint8_t *output;
}
thrdpl_task_t;

typedef struct
{
	const size_t *count;
	int state;
	pthread_mutex_t mutex;
	pthread_cond_t cond;
	pthread_t thread;
	thrdpl_task_t task;
}
thrdpl_thread_t;

typedef struct
{
	size_t thread_count;
	thrdpl_thread_t thread_data[MAX_THREADS];
}
thrdpl_data_t;

// ==========================================================================
// Helper macros
// ==========================================================================

#define BOUND(MIN, VAL, MAX) \
{ \
	if ((VAL) > (MAX)) { VAL = (MAX); } \
	if ((VAL) < (MIN)) { VAL = (MIN); } \
} \
while (0)

#define PTHRD_MUTEX_LOCK(X) do \
{ \
	if (pthread_mutex_lock((X)) != 0) \
	{ \
		abort(); \
	} \
} \
while(0)

#define PTHRD_MUTEX_UNLOCK(X) do \
{ \
	if (pthread_mutex_unlock((X)) != 0) \
	{ \
		abort(); \
	} \
} \
while(0)

#define PTHRD_COND_BROADCAST(X) do \
{ \
	if (pthread_cond_broadcast((X)) != 0) \
	{ \
		abort(); \
	} \
} \
while(0)

#define PTHRD_COND_WAIT(X,Y) do \
{ \
	if (pthread_cond_wait((X), (Y)) != 0) \
	{ \
		abort(); \
	} \
} \
while(0)

#define CHECK_IF_CANCELLED() do \
{ \
	if (data->state == THRD_STATE_EXIT) \
	{ \
		if (pthread_mutex_unlock(&data->mutex) != 0) \
		{ \
			abort(); \
		} \
		return NULL; /* cancelled */ \
	} \
} \
while(0)

// ==========================================================================
// Thread main
// ==========================================================================

static void *worker_thread_main(void *const arg)
{
	thrdpl_thread_t *const data = (thrdpl_thread_t*) arg;
	thrdpl_task_t *task;

	for (;;)
	{
		PTHRD_MUTEX_LOCK(&data->mutex);
		CHECK_IF_CANCELLED();

		while (data->state != THRD_STATE_WORK)
		{
			PTHRD_COND_WAIT(&data->cond, &data->mutex);
			CHECK_IF_CANCELLED();
		}

		task = &data->task;
		PTHRD_MUTEX_UNLOCK(&data->mutex);

		task->worker(*data->count, task->context, task->input, task->output, task->length);

		PTHRD_MUTEX_LOCK(&data->mutex);
		CHECK_IF_CANCELLED();
		data->state = THRD_STATE_IDLE;
		PTHRD_COND_BROADCAST(&data->cond);
		PTHRD_MUTEX_UNLOCK(&data->mutex);
	}
}

// ==========================================================================
// Manage threads
// ==========================================================================

#if defined(__unix__)
#	define GET_NPROCS_FUNCTION() get_nprocs()
#elif defined(PTW32_VERSION)
#	define GET_NPROCS_FUNCTION() pthread_num_processors_np()
#endif

static size_t detect_cpu_count(void)
{
#ifdef GET_NPROCS_FUNCTION
	const int cpu_count = GET_NPROCS_FUNCTION();
	if (cpu_count > 0)
	{
		return (size_t) cpu_count;
	}
#endif
	return 1U;
}

static int create_worker_thread(thrdpl_thread_t *const thread_data, const size_t *const count)
{
	thread_data->count = count;
	thread_data->state = THRD_STATE_IDLE;

	if (pthread_mutex_init(&thread_data->mutex, NULL) != 0)
	{
		return -1;
	}

	if (pthread_cond_init(&thread_data->cond, NULL) != 0)
	{
		pthread_mutex_destroy(&thread_data->mutex);
		return -1;
	}

	if (pthread_create(&thread_data->thread, NULL, worker_thread_main, thread_data) != 0)
	{
		pthread_cond_destroy(&thread_data->cond);
		pthread_mutex_destroy(&thread_data->mutex);
		return -1;
	}

	return 0;
}

static int destroy_worker_thread(thrdpl_thread_t *const thread_data)
{
	PTHRD_MUTEX_LOCK(&thread_data->mutex);
	thread_data->state = THRD_STATE_EXIT;
	PTHRD_COND_BROADCAST(&thread_data->cond);
	PTHRD_MUTEX_UNLOCK(&thread_data->mutex);

	pthread_join(thread_data->thread, NULL);
	pthread_mutex_destroy(&thread_data->mutex);
	pthread_cond_destroy(&thread_data->cond);

	return 0;
}

// ==========================================================================
// Thread pool API
// ==========================================================================

thrdpl_t slunkcrypt_thrdpl_create(const size_t count)
{
	size_t i, j;
	thrdpl_data_t *pool = NULL;

	if (!(pool = (thrdpl_data_t*) malloc(sizeof(thrdpl_data_t))))
	{
		return THRDPL_NULL;
	}

	slunkcrypt_bzero(pool, sizeof(thrdpl_data_t));
	pool->thread_count = (count > 0U) ? count : detect_cpu_count();
	BOUND(MIN_THREADS, pool->thread_count, MAX_THREADS);

	for (i = 0U; i < pool->thread_count; ++i)
	{
		if (create_worker_thread(&pool->thread_data[i], &pool->thread_count) != 0)
		{
			for (j = 0U; j < i; ++j)
			{
				destroy_worker_thread(&pool->thread_data[j]);
			}
			goto failure;
		}
	}

	return (thrdpl_t)pool;

failure:
	free(pool);
	return (thrdpl_t)NULL;
}

size_t slunkcrypt_thrdpl_count(const thrdpl_t thrdpl)
{
	thrdpl_data_t *const pool = (thrdpl_data_t*) thrdpl;
	return pool->thread_count;
}

void slunkcrypt_thrdpl_exec(const thrdpl_t thrdpl, const size_t index, const thrdpl_worker_t worker, void *const context, const uint8_t *const input, uint8_t *const output, const size_t length)
{
	thrdpl_data_t *const pool = (thrdpl_data_t*) thrdpl;
	thrdpl_thread_t *const thread = &pool->thread_data[index];

	PTHRD_MUTEX_LOCK(&thread->mutex);

	while ((thread->state != THRD_STATE_IDLE) && (thread->state != THRD_STATE_EXIT))
	{
		if (pthread_cond_wait(&thread->cond, &thread->mutex) != 0)
		{
			abort();
		}
	}

	if (thread->state == THRD_STATE_EXIT)
	{
		abort(); /*this is not supposed to happen!*/
	}

	thread->task.worker = worker;
	thread->task.context = context;
	thread->task.input = input;
	thread->task.output = output;
	thread->task.length = length;
	thread->state = THRD_STATE_WORK;

	PTHRD_COND_BROADCAST(&thread->cond);
	PTHRD_MUTEX_UNLOCK(&thread->mutex);
}

void slunkcrypt_thrdpl_await(const thrdpl_t thrdpl)
{
	size_t i;
	thrdpl_data_t *const pool = (thrdpl_data_t*) thrdpl;

	for (i = 0; i < pool->thread_count; ++i)
	{
		if (pthread_mutex_lock(&pool->thread_data[i].mutex) != 0)
		{
			abort();
		}
		while ((pool->thread_data[i].state != THRD_STATE_IDLE) && (pool->thread_data[i].state != THRD_STATE_EXIT))
		{
			if (pthread_cond_wait(&pool->thread_data[i].cond, &pool->thread_data[i].mutex) != 0)
			{
				abort();
			}
		}
		if (pthread_mutex_unlock(&pool->thread_data[i].mutex) != 0)
		{
			abort();
		}
	}
}

void slunkcrypt_thrdpl_destroy(const thrdpl_t thrdpl)
{
	size_t i;
	thrdpl_data_t *const pool = (thrdpl_data_t*) thrdpl;
	if (pool)
	{
		for (i = 0U; i < pool->thread_count; ++i)
		{
			destroy_worker_thread(&pool->thread_data[i]);
		}
		slunkcrypt_bzero(pool, sizeof(thrdpl_data_t));
		free(pool);
	}
}
